#include "stereo.h"
#include <memory.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <mex.h>

#define SAFE

#define DEFAULT_GAMMA 10

#define DISPARITY(m) ((int16)((m).left) - (int16)((m).right))
#define D_MAX (dps.dMax)
#define EPI_STEP(epi) (((epi) % 2) ? 1 : -1)
#define OCCLUSION 255

// Assumes byte-sized but possibly multiband pixels.
#define IMAGE_BYTE(ip, r, c, b) ((uint8 *)FORTRAN_PIXEL_POINTER(ip, r, c))[b]
#define FIRST_IMAGE_ROW 0
#define FIRST_IMAGE_COL 0
#define LAST_IMAGE_ROW dps.lastRow
#define LAST_IMAGE_COL dps.lastCol

#define LEFT 0
#define RIGHT 1

#define BLOCK_ENTRY(which, m) FORTRAN_BLOCK_POINTER(which, (m).epi, (m).left, DISPARITY(m))

#define COST_STATE(m) ((uint32 *)FORTRAN_PIXEL_POINTER(dps.cumCost, (m).left, DISPARITY(m)))[0]
#define HBACK_STATE(m) ((uint8 *)FORTRAN_PIXEL_POINTER(dps.back, (m).left, DISPARITY(m)))[0]
#define MATCH_COST(m) (*((uint8 *)FORTRAN_BLOCK_POINTER(dps.matchCost, (m).epi, (m).left, DISPARITY(m))))

#define INVALID 0xffff

#ifndef MAX
#define MAX(a, b) (((a) > (b)) ? (a) : (b))
#endif

#ifndef MIN
#define MIN(a, b) (((a) < (b)) ? (a) : (b))
#endif

typedef struct
{
	uint16 epi, left, right;
} match;

typedef struct
{
	match m[3];
	uint8 n;
} predList;

typedef struct
{
	uint16 dMax;
	IMAGEP back;
	IMAGEP matchCost, cumCost;
	uint16 lastRow, lastCol;
	uint32 gamma;
} dpStorage;

static dpStorage dps;

// Image is assumed to be indexed [0, m-1] x [0, n-1].
// DP 'back' structure is [0, m-1] x [0, n-1] x [0, dmax]

inline uint32 getCostState(match &m)
{
	return COST_STATE(m);
}

inline void setCostState(match &m, uint32 cost)
{
	COST_STATE(m) = cost;
}

void dpInit(IMAGEP cost, uint16 dMax)
{
	dps.gamma = DEFAULT_GAMMA;
	dps.dMax = dMax;
	dps.matchCost = cost;
	dps.cumCost = newImage(cost->cols, cost->bands, 1, 4, t, f);
	dps.back = newImage(cost->rows, cost->cols, cost->bands, 1, t, f);
	dps.lastRow = dps.matchCost->rows - 1;
	dps.lastCol = dps.matchCost->cols - 1;
}

void dpFree()
{
	freeImage(&(dps.back));
}

inline void invalidate(match& m)
{
	m.epi = INVALID;
	m.left = INVALID;
	m.right = INVALID;
}

inline logical isInvalid(match& m)
{
	return (m.epi == INVALID || m.left == INVALID || m.right == INVALID) ? t : f;
}

inline logical isInBounds(match& m)
{
	if (isInvalid(m)) return f;

	int16 disparity = DISPARITY(m);

	return (m.epi >= FIRST_IMAGE_ROW && m.epi <= LAST_IMAGE_ROW &&
		m.left >= FIRST_IMAGE_COL && m.left <= LAST_IMAGE_COL &&
		m.right >= FIRST_IMAGE_COL && m.right <= LAST_IMAGE_COL &&
		disparity >= 0 && disparity <= D_MAX) ? t : f;
}

inline void next(match& m)
{
	if (!isInBounds(m))
	{
		invalidate(m);
		return;
	}
	
	int16 disparity = DISPARITY(m);
	
	if (m.left < LAST_IMAGE_COL && disparity < D_MAX)
		// Next disparity
	{
		m.left++;
	}
	else
	{
#ifdef SAFE
		if (disparity == D_MAX || m.left == LAST_IMAGE_COL)
#endif
			// Next pixel, or stop
		{
			if (m.right < LAST_IMAGE_COL)
			{
				m.left = m.right + 1;
				m.right = m.left;
			}
			else
			{
				invalidate(m);
			}
		}
#ifdef SAFE
		else
		{
			tell(t, "next got to where it shouldn't (left-to-right) with m = (%d, %d, %d)",
				(int)m.epi, (int)m.left, (int)m.right);
		}
#endif
	}
}

inline logical restart(match m)
{
	if (m.left == FIRST_IMAGE_COL && m.right == FIRST_IMAGE_COL) return t;

	return f;
}

inline void code(int8 step[], uint8& byte)
{
	uint8 b = 0;

	if (step[0] != 0 || step[1] != 0) // Zero motion is encoded as zero in any case
	{
		b = step[0] + (step[1] << 1);
	}

	// Clear the relevant bits in byte
	uint8 mask = ~(0x3);
	byte &= mask;

	// Store the code
	byte |= b;
}

inline void decode(uint8 byte, int8 step[])
{
	byte = byte & 0x3;

	// Zero motion
	if (byte == 0)
	{
		step[0] = step[1] = 0;
		return;
	}

	step[0] = (byte & 0x1) ? 1 : 0;
	step[1] = (byte & 0x2) ? 1 : 0;
}

inline void back(match& m)
{
	if (!isInBounds(m))
	{
		invalidate(m);
		return;
	}

	int8 step[2];

	decode(HBACK_STATE(m), step);

	if (step[0] == 0 && step[1] == 0)	// No motion
	{
		invalidate(m);
	}
	else
	{
		m.left -= step[0];
		m.right -= step[1];
	}
}

inline void setBack(match& m, int8 step[2])
{
	uint8 byte = 0, mask = 0;

	code(step, byte);
	mask = 0x3;
	HBACK_STATE(m) &= ~mask;	// Erase bits in the relevant slice
	HBACK_STATE(m) |= (byte & mask);  // Write new bits
}

inline void pred(match m, predList& list)
{
	match p;
	int8 lStep[3], rStep[3];

	list.n = 0;

	lStep[0] = 1; rStep[0] = 0;
	lStep[1] = 1; rStep[1] = 1;
	lStep[2] = 0; rStep[2] = 1;

	for (uint8 k = 0; k < 3; k++)
	{
		p.epi = m.epi;
		p.left = m.left - lStep[k];
		p.right = m.right - rStep[k];
		if (isInBounds(p))
		{
			list.m[list.n].epi = p.epi;
			list.m[list.n].left = p.left;
			list.m[list.n].right = p.right;
			list.n++;
		}
	}
}

inline uint32 stepCost(match& p, match& m)
{
	// We first list all cases in which there is a cost discount.
	// If all those fail, a standard step cost is levied at the bottom.

	if (isInvalid(m) || isInvalid(p))
	{
		return (uint32)0;
	}

	if (m.left == LAST_IMAGE_COL || m.right == FIRST_IMAGE_COL)
	{
		return 0;		// Free occlusions at image boundaries
	}
	
	if (abs((int)m.left - (int)p.left) == 1 && abs((int)m.right - p.right) == 1)
	{
		return 0;		// No occlusion, no penalty
	}

	return 8;		// Maximum penalty for occlusions everywhere else
}

inline uint32 differenceCost(match& m)
{
	if (!isInBounds(m))
	{
		return (uint32)0;
	}
	else
	{
		return (uint32)MATCH_COST(m);
	}
}

void inline start(uint16 k, match& m, int8 direction)
{
	m.epi = k;
	if (direction > 0)
	{
		m.left = FIRST_IMAGE_COL;
		m.right = FIRST_IMAGE_COL;
	}
	else	// Right-to-left
	{
		m.left = LAST_IMAGE_COL;
		m.right = LAST_IMAGE_COL;
	}
}

inline void update(uint16 k)
{
	uint32 cost, diffCost;
	predList p;
	int8 step[3];
	match m;

	for (start(k, m, 1); isInBounds(m); next(m))
	{
		diffCost = differenceCost(m);

		if (restart(m))
		{
			/*
			logical beginning = (m.epi == FIRST_IMAGE_ROW) ? t : f;

			if (beginning)
			{
				cost = diffCost;
			}
			else
			{
				cost = 0xffffffff;
			}
			*/
			cost = diffCost;
			step[0] = step[1] = 0;
		}
		else
		{
			pred(m, p);

			if (p.n > 0)
			{
				uint32 c, cs, sc, minCost = 0xffffffff;
				uint8 hMin = 0;
				for (uint8 k = 0; k < p.n; k++)
				{
					uint8 h = k;

					cs = getCostState(p.m[h]);
					sc = stepCost(p.m[h], m);

					/*
					if (m.left == FIRST_IMAGE_ROW)
					{
						c = diffCost;
					}
					else if (m.right == LAST_IMAGE_ROW)
					{
						c = cs;
					}
					else*/
					{
						c = cs + sc + diffCost;
					}

					if (c < minCost)
					{
						minCost = c;
						hMin = h;
					}
				}
				cost = minCost;
				step[0] = m.left - p.m[hMin].left;
				step[1] = m.right - p.m[hMin].right;
			}
			else
			{
				cost = diffCost;
				step[0] = step[1] = 0;
			}
		}

		// Save the state
		setCostState(m, cost);
		setBack(m, step);
	}
}

void backTrack(uint16 k, IMAGEP leftDepth, IMAGEP rightDepth)
{
	match m, p;

	invalidate(p);
	for (start(k, m, -1); ; )
	{
		uint8 disparity = (uint8)DISPARITY(m);
		
		if (isInvalid(p))
		{
			IMAGE_BYTE(leftDepth, m.epi, m.left, 0) = disparity;
			IMAGE_BYTE(rightDepth, m.epi, m.right, 0) = disparity;
		}
		else
		{
			// No occlusion
			if (p.right != m.right && p.left != m.left)
			{
				IMAGE_BYTE(leftDepth, m.epi, m.left, 0) = disparity;
				IMAGE_BYTE(rightDepth, m.epi, m.right, 0) = disparity;
			}
			else if (p.left == m.left)	// Left occlusion
			{
				IMAGE_BYTE(rightDepth, m.epi, m.right, 0) = OCCLUSION;
			}
			else if (p.right == m.right)	// Right occlusion
			{
				IMAGE_BYTE(leftDepth, p.epi, p.left, 0) = OCCLUSION;
				IMAGE_BYTE(leftDepth, m.epi, m.left, 0) = disparity;
				IMAGE_BYTE(rightDepth, m.epi, m.right, 0) = disparity;
			}
		}
		
		p.epi = m.epi;
		p.left = m.left;
		p.right = m.right;
		
		back(m);
		
		if (!isInBounds(m)) break;
	}
}

void dpStereo(IMAGEP cost, uint16 dMax, IMAGEP leftDepth, IMAGEP rightDepth)
{

	/*
	match m;
	int8 d, de;
	uint8 r, c, b, v, w;
	*/

	dpInit(cost, dMax);

	/*
	// This loop is for debugging
	for (r = 0; r < cost->rows; r++)
	{
		for (c = 0; c < cost->cols; c++)
		{
			for (b = 0; b < cost->bands; b++)
			{
				v = *((uint8 *)FORTRAN_BLOCK_POINTER(cost, r, c, b));
				w = *((uint8 *)FORTRAN_BLOCK_POINTER(dps.matchCost, r, c, b));
			}
		}
	}
	*/

	for (uint16 epi = FIRST_IMAGE_ROW; epi <= LAST_IMAGE_ROW; epi++)
	{
		update(epi);
		backTrack(epi, leftDepth, rightDepth);
	}

	dpFree();
}